# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/cif2lmdb.ipynb (unless otherwise specified).

__all__ = ['get_instance', 'update_instance_data', 'set_instance_data', 'map_size', 'convert', 'convert_cmd']

# Cell
import math
import javabridge
import bioformats as bf
import numpy as np
import logging
from tqdm import tqdm
from os.path import basename, splitext, join, exists
from os import mkdir
from shutil import rmtree
import lmdb
import pickle
from pathlib import Path
import numpy
import click

# Internal Cell

def _is_image(_series, _r):
    _r.rdr.setSeries(_series)
    return _r.rdr.getPixelType() != 1

# Cell
def get_instance(shape, num_channels):
    return [
        np.empty([num_channels] + list(shape), dtype=np.uint16),
        np.empty([num_channels] + list(shape), dtype=np.bool)
    ]

def update_instance_data(instance, channel, img, mask):
    instance[0][channel] = img
    instance[1][channel] = mask
    return instance

def set_instance_data(instance, img, mask):
    instance[0] = img
    instance[1] = mask
    return instance

map_size = 1099511627776 # 1 TiB

# Cell

def convert(cif, output, channels, names, debug, overwrite, targets_npy, skip_npy):

    if debug:
        logging.basicConfig()
        logging.getLogger().setLevel(logging.DEBUG)

    if output is None:
        output = cif.replace(".cif", ".lmdb")

    if overwrite and exists(output):
        Path(output).unlink()
        if Path(output+"-lock").exists():
            Path(output + "-lock").unlink()
    elif not overwrite and exists(output):
        raise ValueError(output, "Output path exists.")

    logger = logging.getLogger(__name__)

    try:
        logger.debug("Starting Java VM")
        javabridge.start_vm(class_path=bf.JARS, run_headless=True, max_heap_size="8G")
        logger.debug("Started Java VM")

        reader = bf.formatreader.get_image_reader("tmp", path=cif)
        r_length = javabridge.call(reader.metadata, "getImageCount", "()I")
        num_channels = javabridge.call(reader.metadata, "getChannelCount", "(I)I", 0)

        if debug:
            r_length=200

        if len(channels) == 0:
            crange = [i for i in range(num_channels)]
        else:
            crange = np.array(channels)-1

        if len(names) != 0 and len(crange) != len(names):
            raise ValueError(names, "Incorrect amount of names for channels.")

        idx_bytes = int(np.ceil((np.floor(np.log2(r_length))+1)/8.))

        env = lmdb.open(output, lock=False, map_size=map_size, subdir=False, metasync=False, meminit=False, sync=False, map_async=True)
        try:
            logger.info("Opening lmdb database %s" % output)
            output_length = r_length//2

            with env.begin(write=True) as txn:

                skip = []
                if targets_npy is not None:
                    # write some metadata
                    targets = list(numpy.load(targets_npy))

                    if skip_npy is not None:
                        skip = list(numpy.load(skip_npy))
                        output_length -= len(skip)

                    if not debug: # skip test when debugging
                        assert len(targets) == output_length, "Number of targets does not match eventual lmdb length (meta %d, lmdb %d)." % (len(targets), output_length)

                    txn.put(b'__targets__', pickle.dumps(targets))

                txn.put(b'__len__', (output_length).to_bytes(idx_bytes, "big"))
                if names is not None:
                    txn.put(b'__names__', " ".join(names).encode("utf-8"))

            bar = tqdm(range(r_length)[::2])
            skip_count = 0
            counter = 0
            for i in bar:
                if i//2 in skip:
                    skip_count += 1
                    bar.set_postfix({"#skipped": skip_count})
                    continue

                instance = None
                for c_idx, c in enumerate(crange):
                    image = reader.read(c=c, series=i, rescale=False)
                    mask = reader.read(c=c, series=i+1, rescale=False)
                    if instance == None:
                        instance = get_instance(image.shape, len(crange))
                    update_instance_data(instance, c_idx, image, mask)

                with env.begin(write=True) as txn:
                    txn.put(counter.to_bytes(idx_bytes, byteorder='big'), pickle.dumps(instance))

                counter += 1
        finally:
            env.sync()
            env.close()
    finally:
        javabridge.kill_vm()

# Cell
@click.command(name="cif2lmdb")
@click.argument("cif", type=click.Path(exists=True, dir_okay=False))
@click.option("--output", type=click.Path(exists=False, dir_okay=False), default=None, help="Output filename. If not set, cif-filename is taken with lmdb extension.")
@click.option("--channels", multiple=True, type=int, default=[], help="Images from these channels will be extracted. Default is to extract all. 1-based index.")
@click.option("--names", multiple=True, type=str, help="Names to assign to channels.", default=[])
@click.option("--debug", is_flag=True, flag_value=True, help="Show debugging information. Limits output to 100 first cells.", default=False)
@click.option("--overwrite", is_flag=True, flag_value=True, help="Overwrite lmdb if it exists.", default=False)
@click.option("--targets-npy", type=click.Path(exists=True, dir_okay=False), help="Numpy binary file containing targets.", default=None)
@click.option("--skip-npy", type=click.Path(exists=True, dir_okay=False), help="Numpy binary file containing instances to be skipped.", default=None)
def convert_cmd(cif, output, channels, names, debug, overwrite, targets_npy, skip_npy):
    convert(cif, output, channels, names, debug, overwrite, targets_npy, skip_npy)


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/utils.ipynb (unless otherwise specified).

__all__ = ['deep_copy_lmdb']

# Cell
from ifcimglib import imglmdb
import lmdb
import os

# Cell
def deep_copy_lmdb(old_lmdb_env, new_name):
    fd = os.open(new_name, os.O_RDWR)
    old_lmdb_env.copyfd(fd)
    os.close(fd)


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/imglmdb.ipynb (unless otherwise specified).

__all__ = ['imglmdb', 'multidbwrapper']

# Cell
import lmdb
import os.path
import pickle
import numpy
import logging

# Cell
class imglmdb:

    def __init__(self, db_path, endianess="big"):
        self.db_path = db_path
        self.env = lmdb.open(db_path, subdir=os.path.isdir(db_path),
            readonly=True, lock=False,readahead=False, meminit=False)
        self.endianess = endianess
        self.logger = logging.getLogger(__name__)
        self.dropped = set()

        with self.get_read_txn() as txn:
            self.length = int.from_bytes(txn.get(b"__len__"), endianess)
            self.names = txn.get(b'__names__').decode("utf-8").split(' ')
            self.channels_of_interest = [i for i in range(len(self.names))]

            self.targets = txn.get(b'__targets__', default=None)
            if self.targets is not None:
                self.targets = pickle.loads(targets)
        self.idx_byte_length = int(numpy.ceil(numpy.floor(numpy.log2(self.length))/8.))

        self.logger.info("Initialized db (%s) with length %d, and channel names %s" % (self.db_path, self.length, " ".join(self.names)))

    def set_channels_of_interest(self, arr):
        self.channels_of_interest = arr
        self.logger.info("Channels of interest set to %s" % " ".join([self.names[i] for i in arr]))

    def get_read_txn(self):
        return self.env.begin(write=False)

    def get_write_txn(self):
        return self.env.begin(write=True)

    def get_masked_image(self, idx, only_coi = False, txn = None):
        ret = self.get_image(idx, only_coi, txn)

        if len(ret) > 2:
            return numpy.multiply(ret[0], ret[1]), ret[2]
        else:
            return numpy.multiply(ret[0], ret[1])

    def get_image(self, idx, only_coi = False, txn = None):

        if idx in self.dropped:
            raise ValueError("Index is dropped. Call `reset` to reset dropped elements.")

        if txn is None:
            with self.get_read_txn() as txn:
                i, m = pickle.loads(txn.get(int(idx).to_bytes(self.idx_byte_length, "big")))
        else:
            i, m = pickle.loads(txn.get(int(idx).to_bytes(self.idx_byte_length, "big")))

        if only_coi:
            i = i[self.channels_of_interest]
            m = m[self.channels_of_interest]

        if self.targets is None:
            return numpy.array(i), numpy.array(m)

        return numpy.array(i), numpy.array(m), self.targets[idx]

    def get_images(self, idx, only_coi = False, only_image=False, masked=False):
        func = self.get_masked_image if masked else self.get_image

        with self.get_read_txn() as txn:
            ret = []
            if only_image:
                for i in idx:
                    try:
                        ret.append(func(i, only_coi, txn)[0])
                    except ValueError:
                        pass
            else:
                for i in idx:
                    try:
                        ret.append(func(i, only_coi, txn))
                    except ValueError:
                        pass

            return ret


    def __iter__(self):
        self.pointer = 0
        return self

    def __next__(self):
        if self.pointer < self.length:
            while self.pointer in self.dropped: # skip virtually dropped instances
                self.pointer += 1
            self.pointer += 1
            return self.get_image(self.pointer-1)
        else:
            raise StopIteration()

    def drop(self, idx):
        self.dropped.update(idx)

    def drop_commit(self):
        with self.get_write_txn() as txn:
            for idx in self.dropped:
                txn.delete(
                    int(idx).to_bytes(self.idx_byte_length, "big")
                )

    def reset(self):
        self.dropped = set()

    def __del__(self):
        self.env.close()

    def __len__(self):
        return self.length - len(self.dropped)

    def __repr__(self):
        return "db %s, length %d, channels %s" % (self.db_path.split("/")[-1], self.length, " ".join(self.names))

# Cell

class multidbwrapper:

    def __init__(self, db_paths, channels=[]):
        self.db_paths = db_paths
        self.dbs = []
        self.db_start_index = []

        self.length = 0
        tmp_channels = []
        self.__targets = []
        for db_path in self.db_paths:
            db = imglmdb(db_path)
            self.length += len(db)
            tmp_channels.append(db.names)
            self.__targets.extend(db.targets)

        self.__targets = numpy.array(self.__targets)
        self.__label_offset = self.__targets.min()
        self.__targets -= self.__label_offset
        self.__classes = numpy.unique(self.__targets)

        if all(len(i) == len(tmp_channels[0]) for i in tmp_channels):
            if len(channels) > 0:
                self.__channels_of_interest = channels
            else:
                self.__channels_of_interest = [i for i in range(len(tmp_channels[0]))]
        else:
            raise ValueError("Not all DBs contain the same amount of channels.")

    @property
    def targets(self):
        return self.__targets
    @property
    def classes(self):
        return self.__classes
    @property
    def dtype(self):
        return numpy.float32
    @property
    def channels_of_interest(self):
        return self.__channels_of_interest
    @property
    def label_offset(self):
        return self.__label_offset
    def __len__(self):
        return self.length
    def __str__(self):
        return f"{self.length} instances from {len(self.dbs)} databases."

    def _setup(self):
        """Private function used to setup databases. Required for multiprocessing.
        """
        i = 0
        for db_path in self.db_paths:
            db = imglmdb(db_path)
            db.set_channels_of_interest(self.channels_of_interest)

            self.dbs.append(db)
            self.db_start_index.append(i)

            i += len(db)

        self.db_start_index = numpy.array(self.db_start_index)

    def get_image(self, index, only_coi=True):
        """Fetches instance from database based on index.

        Arguments:
            index {int} -- Index of instance to be fetched

        Returns:
            tuple -- image [, label]
        """
        if len(self.dbs) == 0:
            self._setup()

        db_idx = sum(self.db_start_index - index <= 0)-1
        db = self.dbs[db_idx]
        start_idx = self.db_start_index[db_idx]

        image, mask, label = db.get_image(index-start_idx, only_coi=only_coi)

        return image, mask, label-self.label_offset


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/apply.ipynb (unless otherwise specified).

__all__ = ['preprocess_db_intra_image', 'aggregate_fold_stats']

# Cell
from ifcimglib import imglmdb, utils, preprocessing, cif2lmdb
import numpy
import matplotlib.pyplot as plt
from tqdm import trange
import pickle
import logging
import lmdb
from pathlib import Path
from tqdm import tqdm
import os
import seaborn
from sklearn.model_selection import PredefinedSplit

# Cell

def preprocess_db_intra_image(db, preprocessed_output_path):
    logger = logging.getLogger(__name__)

    if Path(preprocessed_output_path).exists():
        Path(preprocessed_output_path).unlink()

    env = lmdb.open(preprocessed_output_path, lock=False, sync=False, map_size=cif2lmdb.map_size, subdir=False)
    logger.info("Opened lmdb database %s" % preprocessed_output_path)

    with env.begin(write=True) as txn:
        txn.put(b'__targets__', pickle.dumps(db.targets))
        txn.put(b'__len__', len(db).to_bytes(db.idx_byte_length, "big"))
        txn.put(b'__names__', " ".join(db.names).encode("utf-8"))

        for i in trange(len(db)):
            x, m, _ = db.get_image(i)
            x = x.astype(numpy.float32)
            x = preprocessing.log_transform(x, m, [1])
            x = preprocessing.min_max_normalize(x, m, "clip")
            x = preprocessing.crop_and_pad_to_square(x, 70)
            m = preprocessing.crop_and_pad_to_square(m.astype(numpy.uint8), 70).astype(bool)

            instance = cif2lmdb.get_instance(x.shape[1:], x.shape[0])
            instance = cif2lmdb.set_instance_data(instance, x.astype(numpy.float16), m)

            txn.put(i.to_bytes(db.idx_byte_length, byteorder='big'), pickle.dumps(instance))
    env.sync()
    env.close()

# Cell
def aggregate_fold_stats(db_paths, cv_pkl_file):
    preprocessed_db = imglmdb.multidbwrapper(sorted(db_paths))
    with open(cv_pkl_file, "rb") as pkl:
        test_fold, nested_test_folds = pickle.load(pkl)

    splitter = PredefinedSplit(test_fold)

    data = [{}]*splitter.get_n_splits()

    for i, (nested_test_fold, (_, test_idx)) in enumerate(zip(nested_test_folds, splitter.split())):
        per_pixel_stats = preprocessing.compute_per_pixel_stats(preprocessed_db, None, idx=test_idx)
        std_per_pixel = numpy.where(per_pixel_stats[1] == 0.0, 1, per_pixel_stats[1])
        data[i]["outer"] = (per_pixel_stats[0], std_per_pixel)

        nested_splitter = PredefinedSplit(nested_test_fold)
        data[i]["nested"] = [{}]*nested_splitter.get_n_splits()

        for j, (train_idx, val_idx) in enumerate(nested_splitter.split()):
            per_pixel_stats = preprocessing.compute_per_pixel_stats(preprocessed_db, None, idx=train_idx)
            std_per_pixel = numpy.where(per_pixel_stats[1] == 0.0, 1, per_pixel_stats[1])
            data[i]["nested"][j]["train"] = (per_pixel_stats[0], std_per_pixel)

            per_pixel_stats = preprocessing.compute_per_pixel_stats(preprocessed_db, None, idx=val_idx)
            std_per_pixel = numpy.where(per_pixel_stats[1] == 0.0, 1, per_pixel_stats[1])
            data[i]["nested"][j]["val"] = (per_pixel_stats[0], std_per_pixel)

    with open(os.path.splitext(cv_pkl_file)[0] + "_stats.pkl", "wb") as pkl:
        pickle.dump(data, pkl)

    return data


# AUTOGENERATED! DO NOT EDIT! File to edit: notebooks/cif2tiff.ipynb (unless otherwise specified).

__all__ = ['convert', 'convert_cmd']

# Cell
import math
import javabridge
import bioformats as bf
import numpy as np
import logging
from tqdm import tqdm
from pathlib import Path
import numpy
import click
import flowio
from PIL import Image
from multiprocessing import Pool

# Internal Cell

def _is_image(_series, _r):
    _r.rdr.setSeries(_series)
    return _r.rdr.getPixelType() != 1

# Internal Cell
def setup_directory_structure(out, fcs):
    data = flowio.FlowData(fcs)

    label_idx = None
    for k, v in data.channels.items():
        if v["PnN"] == "label":
            label_idx = int(k)

    assert label_idx is not None, "No label column present in fcs file"

    labels = numpy.reshape(data.events, (-1, data.channel_count))[:, label_idx-1].astype(int)

    out = Path(out)
    for label in numpy.unique(labels):
        (Path(out) / str(label)).mkdir(exist_ok=True, parents=True)

    return labels

# Internal Cell

def process_chunk(images, crange):
    upper_bound = images[0].reshape(len(crange), -1).max(axis=1)
    lower_bound = images[0].reshape(len(crange), -1).min(axis=1)

    counter = 1
    for i in images[1:]:
        a = images[counter].reshape(len(crange), -1).min(axis=1)
        b = images[counter].reshape(len(crange), -1).max(axis=1)
        lower_bound = np.where(a < lower_bound, a, lower_bound)
        upper_bound = np.where(b > upper_bound, b, upper_bound)

        counter += 1

    return lower_bound, upper_bound

def get_min_max_in_file(reader, r_length, crange, nprocs, do_tqdm=True):

    chunks = numpy.array_split(numpy.arange(0, r_length, step=2), nprocs)
    image_chunks = [None]*len(chunks)

    with Pool(processes=nprocs) as pool:
        results = []
        for i, chunk in tqdm(enumerate(chunks), position=0, leave=False, total=len(chunks)):

            images = [None]*len(chunk)
            for j, series in tqdm(enumerate(chunk), total=len(chunk), position=1, leave=False, mininterval=10):
                first = reader.read(c=0, series=series)
                im = numpy.empty(shape=(len(crange),) + first.shape, dtype=first.dtype)
                mask = numpy.empty(shape=(len(crange),) + first.shape, dtype=first.dtype)

                im[0] = first
                for c in crange[1:]:
                    im[c] = reader.read(c=c, series=series)
                for c in crange:
                    mask[c] = reader.read(c=c, series=series+1)
                images[j] = im*mask

            image_chunks[i] = images
            results.append(pool.apply_async(process_chunk, args=(images, crange)))
            print(f"Submitted chunk {i}")

        lower_bound, upper_bound = None, None
        for i, result in enumerate(results):
            print(f"Waiting for result {i}")
            a, b = result.get()

            if lower_bound is None:
                lower_bound = a
            else:
                lower_bound = np.where(a < lower_bound, a, lower_bound)
            if upper_bound is None:
                upper_bound = b
            else:
                upper_bound = np.where(b > upper_bound, b, upper_bound)

    return lower_bound, upper_bound, image_chunks

# Cell

def convert(cif_files, fcs_files, output, channels, debug, nproc=1, external_jvm_control=False):

    if debug:
        logging.basicConfig()
        logging.getLogger().setLevel(logging.INFO)

    output = Path(output)

    logger = logging.getLogger(__name__)

    try:

        if not external_jvm_control:
            logger.debug("Starting Java VM")
            javabridge.start_vm(class_path=bf.JARS, run_headless=True, max_heap_size="12G")
            logger.debug("Started Java VM")

        counter = 0
        for cif, fcs in zip(cif_files, fcs_files):
            print(f"Processing {cif}")

            labels = setup_directory_structure(output, fcs)

            reader = bf.formatreader.get_image_reader("reader", path=cif)
            r_length = javabridge.call(reader.metadata, "getImageCount", "()I")
            num_channels = javabridge.call(reader.metadata, "getChannelCount", "(I)I", 0)

            if debug:
                r_length=100

            if len(channels) == 0:
                crange = [i for i in range(num_channels)]
            else:
                crange = np.array(channels)-1

            lower_bound, upper_bound, image_chunks = get_min_max_in_file(reader, r_length, crange, nproc)
            lower_bound = lower_bound.reshape(len(crange), 1, 1)
            upper_bound = upper_bound.reshape(len(crange), 1, 1)

            a = 0
            for images in tqdm(image_chunks, position=0, leave=False, mininterval=10):
                for im in images:
                    im = (((im - lower_bound) / (upper_bound - lower_bound))*(2**16)).astype(numpy.uint16)
                    pillow_img = Image.fromarray(im[0], mode="I;16")
                    pillow_img.save(
                        output / str(labels[a]) / f"{counter}.tiff",
                        append_images = [Image.fromarray(im[i], mode="I;16") for i in range(1, im.shape[0])],
                        save_all = True
                    )
                    counter+=1
                    a+1

    finally:
        if not external_jvm_control:
            javabridge.kill_vm()

# Cell
@click.command(name="cif2tiff")
@click.argument("cif", type=click.Path(exists=True, file_okay=False))
@click.argument("output", type=click.Path(exists=False, file_okay=False))
@click.option("--channels", multiple=True, type=int, default=[], help="Images from these channels will be extracted. Default is to extract all. 1-based index.")
@click.option("--debug", is_flag=True, flag_value=True, help="Show debugging information. Limits output to 100 first cells.", default=False)
@click.option("--nproc", type=int, default=-1, help="Amount of processes to use.")
def convert_cmd(cif, output, channels, debug, nproc):

    if nproc == -1:
        from multiprocessing import cpu_count
        nproc = cpu_count()

    import glob
    cif_files = glob.glob(str(Path(cif) / "**" / "*.cif"))
    fcs_files = [str(Path(f).with_suffix(".fcs")) for f in cif_files]

    convert(cif_files, fcs_files, output, channels, debug, nproc, do_tqdm=False)